//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Tue Mar 19 20:36:13 PST 1996
// Last Modified: Sun Dec 14 18:41:43 GMT-0800 1997
// Filename:      .../sig/src/control/MidiOutput/MidiOutput.cc
// Syntax:        C++ 
// $Smake:        cc -O -g -c -I../include -I./ %b.cc && rm %b.o
//


#include "MidiOutput.h"
#include <iomanip.h>                                /* for setw() function */

Timer MidiOutput::timer;

//////////////////////////////
//
// MidiOutput::MidiOutput
// 	Default parameters: openAuto = 1
//

MidiOutput::MidiOutput(int aPort, int openAuto) : MidiPort(aPort, openAuto) {
   outputRecordQ = 0;       
   outputBuffer.allowGrowth();
   outputBuffer.setSize(0);
}



//////////////////////////////
//
// MidiOutput::~MidiOutput 
//

MidiOutput::~MidiOutput() {
   // do nothing
}



//////////////////////////////
//
// MidiOutput::cont -- continuous controller command.  Takes two parameters:
//	the controller number (0-127) and the data for that controller,
//	which is also a number between 0 and 127.
//

void MidiOutput::cont(int channel, int contNum, int contValue) {
   send(0xB0 | channel, 0);           // controller command
   send(contNum, 0);                  // data parameter 1: controler number
   send(contValue, 1);                // data parameter 2: controler value
}



//////////////////////////////
//
// MidiOutput::off -- send off notes for all keys on a given channel.
//	If channel is -1 (the default), then send note offs to all notes
//      on all 16 channels.  
//

void MidiOutput::off(int channel) {
   if (channel >= 0 && channel < 16) {
      for (int i=0; i<128; i++) {
         play(channel, i, 0);  
      }
   } else {
      cerr << "Error: invalid MIDI channel for off(): " << channel << endl;
   }
}

// This version of off is newer to the MIDI standard, and may not
// work on older MIDI synthesizers.  Advantage is that it takes
// less time to send.

void MidiOutput::off(void) {
   cont(0, 120, 0);
}



//////////////////////////////
//
// MidiOutput::offAll -- sends notes off to all notes on all channels
//

void MidiOutput::offAll(void) {
   int i, j;
   for (i=0; i<16; i++) {
      for (j=0; j<128; j++) {
         play(i, j, 0);
      }
   }
}
   


//////////////////////////////
//
// MidiOutput::play -- plays a midi note on the specified channel.  If
//	the last parameter (attack velocity) is not specified, then
//	interpret to be a note off command.
//

void MidiOutput::play(int channel, int keyno, int attack) {
   if (channel < 0 || channel > 15) {
      cerr << "Error playing note: channel " << channel 
           << " is out of range." << endl;
      return;
   }
   if (keyno < 0 || keyno > 127) {
      cerr << "Error playing note: key number " << keyno 
           << " is out of range." << endl;
      return;
   }
   if (attack < 0 || attack > 127) {
      cerr << "Error playing note: attack velocity " << attack 
           << " is out of range." << endl;
      return;
   }
 
   send(0x90 | channel, 0);               // channel command
   send(keyno, 0);                        // midi key number
   send(attack, 1);                       // attack velocity, flush
}



//////////////////////////////
//
// MidiOutput::pc -- Patch Change: send a MIDI program change.
//

void MidiOutput::pc(int channel, int timbre) {
   send(0xc0 | channel, 0);           // command byte for program change
   send(timbre, 1);                   // data parameter 1: timbre number
}



//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: send a MIDI pitch bend.
//      Parameters are:
//         1. channel   -- MIDI channel offset from 0.
//         2. mostByte  -- most significant 7 bits (coarse tuning)
//         3. leastByte -- least significant 7 bits (fine tuning)
//

void MidiOutput::pw(int channel, int mostByte, int leastByte) {
   send(0xe0 | channel, 0);           // command byte for pitch wheel
   send(leastByte & 0x7f, 0);         // data parameter 1: least sig. byte
   send(mostByte & 0x7f, 1);          // data parameter 2: most sig. byte
}



//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: 14 bit number given as input
//	range for 14 bit number is 0 to 16383.
//

void MidiOutput::pw(int channel, int tuningData) {
   uchar greaterBits = (tuningData >> 7) & 0x7f;
   uchar lesserBits = tuningData & 0x7f;
   pw(channel, greaterBits, lesserBits);
}
   


//////////////////////////////
//
// MidiOutput::pw -- Pitch Wheel: range between -1 to 1 given as input.
//	Range is then converted to a 14 bit number.
//      +1 = highest value of pitch wheel
//	 0 = rest position of pitch wheel
//	-1 = lowest value of pitch wheel
//

void MidiOutput::pw(int channel, double tuningData) {
   if (tuningData < -1.0 || tuningData > 1.0) {
      cerr << "Error: pitch wheel data is out of range: " << tuningData << endl;
      exit(1);
   }

   int output = (int)((tuningData+1.0)/2.0*16383 + 0.5);
   pw(channel, output);
}



//////////////////////////////
//
// MidiOutput::recordStart
//

void MidiOutput::recordStart(char *filename, int format) {
   if (outputRecordQ) {  // already recording, so close old file
      recordStop();
   }

   outputRecordFile.open(filename, ios::out);
   if (!outputRecordFile) {   // open file failed
      cerr << "Error: cannot open file " << filename << endl;
      outputRecordQ = 0;
   } else {
      outputRecordQ = 1;
   }
   
   if (outputRecordQ) {
      switch (format) {
         case 0:   // ascii
            outputRecordType = 0;
            outputRecordFile <<"; delta time/MIDI output at delta time" << endl;
            break;
         case 1:   // binary
            outputRecordType = 1;
            // record the magic number for binary format
            outputRecordFile << (uchar)0xf8 << (uchar)0xf8 
                       << (uchar)0xf8 << (uchar)0xf8;
            break;
         case 2:   // standard MIDI file, type 0
         default:
            outputRecordType = 2;
            // header stuff to be written here
            break;
      }
   }
   
   lastFlushTime = timer.getTime();
}



//////////////////////////////
// 
// MidiOutput::recordStop
//

void MidiOutput::recordStop(void) {
   if (outputRecordQ) {
      // clear the output buffer which has to be done when outputRecordQ = 1:
      send(0xf8, 1);

      outputRecordQ = 0;
      outputRecordFile.close();
   }
}



//////////////////////////////
//
// MidiOutput::reset -- sends the MIDI command 0xFF which
//	should force the MIDI devices on the other side of the 
//	MIDI cable into their power-on reset condition, clear running
//	status, turn off any sounding notes, set Local Control on, and
//	otherwise clean up the state of things.
//

void MidiOutput::reset(void) {
   send(0xff, 1);
}



//////////////////////////////
//
// MidiOutput::send -- send a byte to the MIDI port.
//     default parameter: flushQ = 0
//

void MidiOutput::send(int byte, int flushQ) {

   outputBuffer.append((uchar*)&byte);

   if (flushQ) {
      if (outputRecordQ) {
         switch (outputRecordType) {
            case 0:   // ascii
               writeOutputAscii();
               break;
            case 1:   // binary
               writeOutputBinary();
               break;
            case 2:   // standard MIDI file type 0
               writeOutputMidifile();
               break;
         }
         lastFlushTime = timer.getTime();  // only keep track if recording
      }

/* the following code would be needed if the buffer were defered to
 * the port instead of local
 */
//      for (i=0; i<outputBuffer.getSize()-1; i++) {
//         rawsend(outputBuffer[i], 0);
//      }

      flush(outputBuffer.getBase(), outputBuffer.getSize());
      outputBuffer.setSize(0);   // no more MIDI data in outputBuffer
   }  
}



//////////////////////////////
//
// MidiOutput::sustain -- set the MIDI sustain continuous controller on or off.
//

void MidiOutput::sustain(int channel, int status) {
   if (status) {  // turn on sustain 
      send(0xb0 | channel, 0);         // command byte for cont controller
      send(0x40, 0);                   // sustain controller
      send(127, 1);                    // data: 0..63 = off, 64..127 = on
   } else {       // turn off sustain
      send(0xb0 | channel, 0);
      send(0x40, 0);         
      send(0, 1);           
   }
}



//////////////////////////////
//
// MidiOutput::sysex -- send a MIDI system exclusive single data byte, or
//	an array of values.
//

void MidiOutput::sysex(int data) {
   send(0xf0, 0);                  // command byte for start of sysex
   send(data, 0);                  // system exclusive data
   send(0xf7, 1);                  // command byte for end of sysex
}


void MidiOutput::sysex(uchar* data, int size) {
   send(0xf0, 0);                        // command byte for start of sysex
   for (int i=0; i<size; i++) {
      send(data[i], 0);                 
   }
   send(0xf7, 1);                        // command byte for end of sysex
}


void MidiOutput::sysex(Array<uchar>& data) {
   send(0xf0, 0);                        // command byte for start of sysex
   for (int i=0; i<data.getSize(); i++) {
      send(data[i], 0);                 
   }
   send(0xf7, 1);                        // command byte for end of sysex
}
   


////////////////////////////////////////////////////////////////////////////////
//
// Private functions:
//    MidiOutput::writeFourBytes         -- write long to file as four bytes
//    MidiOutput::writeOutputAscii       -- record midi out to ascii file
//    MidiOutput::writeOutputBinary      -- record midi out to binary file
//    MidiOutput::writeOutputMidifile    -- record midi out to midi file type 0
//



//////////////////////////////
//
// MidiOutput::writeFourBytes -- writes a four-byte number most 
//	significant byte first.
//

void MidiOutput::writeFourBytes(fstream& outputFile, ulong number) {
   #ifdef OTHEREND
      static uchar byte[4];
      byte[0] = (number >> 24);             // most significant byte
      byte[1] = (number >> 16) & 0xff;
      byte[2] = (number >> 8)  & 0xff;
      byte[3] = number & 0xff;              // least significant byte
   
      outputFile << byte[0];
      outputFile << byte[1];
      outputFile << byte[2];
      outputFile << byte[3];
   #else
      outputFile << number;
   #endif
}



//////////////////////////////
//
// MidiOutput::writeOutputAscii
//

void MidiOutput::writeOutputAscii(void) {
   outputRecordFile << dec << setw(6) << (timer.getTime()-lastFlushTime) <<'\t';
   for (int i=0; i<outputBuffer.getSize()-1; i++) {
      if (outputBuffer[i] & 0x80) {
         outputRecordFile << "0x" << hex << setw(2) 
                    << (short)outputBuffer[i] << ' ';
      } else {
         outputRecordFile << dec << setw(3) << (short)outputBuffer[i] << ' ';
      }
   }

   if (outputBuffer[outputBuffer.getSize()-1] & 0x80) {
      outputRecordFile << "0x" << hex << setw(2) 
                 << (short)outputBuffer[outputBuffer.getSize()-1] << endl;
   } else {
      outputRecordFile << dec << setw(3) 
                 << (short)outputBuffer[outputBuffer.getSize()-1] << endl;
   }
}



//////////////////////////////
//
// MidiOutput::writeOutputBinary
//

void MidiOutput::writeOutputBinary(void) {
   if (outputBuffer.getSize() == 1 && outputBuffer[0] == 0xf8) return;

   // write the delta time (four bytes)
   writeFourBytes(outputRecordFile, (ulong)(timer.getTime() - lastFlushTime));

   // write midi data from outputBuffer
   for (int i=0; i<outputBuffer.getSize(); i++) {
      // don't store 0xf8 command since it will be 
      // used to mark the end of the delta time data.
      if (outputBuffer[i] == 0xf8) continue;
      outputRecordFile << outputBuffer[i];
   }
}



//////////////////////////////
//
// MidiOutput::writeOutputMidifile
//

void MidiOutput::writeOutputMidifile(void) {
   // not yet implemented
}


