//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Fri Nov  1 23:35:02 PST 1996
// Last Modified: Thu May  1 12:21:20 GMT-0800 1997
// Filename:      .../sig/src/Generator/SoundFileIn/SoundFileIn.cc
// Syntax:        C++ 
// $Smake:        cc -Wall -g -c %b.cc -I../../../include && rm -f %b.o
//

#include <stdlib.h>
#include <iostream.h>
#include <string.h>
#include "SoundFileIn.h"


//////////////////////////////
//
// SoundFileIn::SoundFileIn
//    defaults: aFileName = "", start = NONE, dur = NONE
//

SoundFileIn::SoundFileIn(char* aFileName, float start, float dur) {
   setName("SoundFileIn");

   filename = NULL;
   maxSampleIndex = -1;
   curSampleIndex = 0;
   begSampleIndex = 0;
   
   if (strlen(aFileName) > 0) {
      setFile(aFileName, start, dur);
      // set the size of the output array to the number of channels
      Output.setSize(sndHeader.getChannels());
      Output.allowGrowth(FALSE);
      Output.zero();
   } else {
      Output.setSize(4);
      Output.allowGrowth(TRUE);
      Output.zero();
   }

   runningQ = TRUE;
   loopingQ = FALSE;
  
   brandname = GENERATOR;

   determineSampleReadingFunction();
}



//////////////////////////////
//
// SoundFileIn::~SoundFileIn
//

SoundFileIn::~SoundFileIn() { 
   delete [] filename;
}



//////////////////////////////
//
// SoundFileIn::action
//

void SoundFileIn::action(void) {
   if (!runningQ) return;

   // read the soundfile samples into the output array

   if (curSampleIndex < 0 || curSampleIndex > maxSampleIndex) {
      Output.zero();
      return;
   }

   fillOutputArray();

   curSampleIndex++;
   if (curSampleIndex > endSampleIndex && loopingQ) {
      curSampleIndex = begSampleIndex;
   }

}



//////////////////////////////
//
// SoundFileIn::getChannels
//

int SoundFileIn::getChannels(void) {
   return sndHeader.getChannels();
}



//////////////////////////////
//
// SoundFileIn::getHeader

SoundHeader& SoundFileIn::getHeader(void) {
   return sndHeader;
}



//////////////////////////////
//
// SoundFileIn::getSamples
//

long SoundFileIn::getSamples(void) {
   return maxSampleIndex + 1;
}



//////////////////////////////
//
// SoundFileIn::output
//     default: channel = 0
//

sampleType SoundFileIn::output(int channel) {
   if (channel < 0 || channel >= sndHeader.getChannels()) 
      return Output.sum();
   else
      return Output[channel];
}



//////////////////////////////
//
// SoundFileIn::printState
//

void SoundFileIn::printState(void) {
   int oldFlag = sndHeader.setOutputType(TYPE_TEXT);
   cerr << "Filename: " << filename << endl;
   cerr << sndHeader;
   sndHeader.setOutputType(oldFlag);
}



//////////////////////////////
//
// SoundFileIn::reset
//     default: sampleIndex = 0
//

void SoundFileIn::reset(int sampleIndex) {
   if (sampleIndex < 0) {
      sndfile.seekp(sndHeader.getDataByteOffset()-1); 
   } else {
      sndfile.seekp((int)(sndHeader.getDataByteOffset() - 1 + 
         (sampleIndex + begSampleIndex) * sndHeader.getChannels() *
          sndHeader.getBitsPerSample() / 8.0));
   }
}



//////////////////////////////
//
// SoundFileIn::setFile
//

void SoundFileIn::setFile(char* aFileName, float start, float dur) {
   if (filename != NULL) delete [] filename;
   filename = new char[strlen(aFileName)+1];
   strcpy(filename, aFileName);

   sndHeader.setHeader(filename);

   if (sndfile.is_open()) sndfile.close();
   sndfile.open(filename, ios::in | ios::bin);
   if(!sndfile.is_open()) {
      cerr << "Error: sound file " << filename 
           << " could not be opened!" << endl;
      exit(1);
   }

   // determine the beginning and ending samples according to start and dur
   maxSampleIndex = sndHeader.getSamples() - 1; 
   if (maxSampleIndex < 0) {
      cerr << "Error: sound file " << filename << " has no data!" << endl;
      exit(1);
   }
   if (start >= 0) {
      begSampleIndex = (long)(start * sndHeader.getSrate());
   } else {
      begSampleIndex = 0;
   }
   if (begSampleIndex > maxSampleIndex) begSampleIndex = 0;
   if (dur >= 0) {
      endSampleIndex = begSampleIndex + (long)(dur * sndHeader.getSrate());
   } else {
      endSampleIndex = maxSampleIndex;
   }
   if (endSampleIndex > maxSampleIndex) {
      endSampleIndex = maxSampleIndex;
   }
 
   curSampleIndex = begSampleIndex;

   // set the file to the current sample
   
   sndfile.seekp((long)(sndHeader.getDataByteOffset() + curSampleIndex * 
      sndHeader.getChannels() * sndHeader.getBitsPerSample() / 8.0));

}



///////////////////////////////////////////////////////////////////////////
//
// private functions
//



//////////////////////////////
//
// SoundFileIn::determineSampleReadingFunction
//

void SoundFileIn::determineSampleReadingFunction(void) {
   switch (sndHeader.getInputType()) {
      case TYPE_SND:                        // big endian samples
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               fillOutputArray = SoundFileIn::fillOutputArray8M_B;
               break;
            case SND_FORMAT_LINEAR_8:
               fillOutputArray = SoundFileIn::fillOutputArray8L_B;
               break;
            case SND_FORMAT_LINEAR_16:
               fillOutputArray = SoundFileIn::fillOutputArray16L_B;
               break;
            case SND_FORMAT_LINEAR_24:
               fillOutputArray = SoundFileIn::fillOutputArray24L_B;
               break;
            case SND_FORMAT_LINEAR_32:
               fillOutputArray = SoundFileIn::fillOutputArray32L_B;
               break;
            case SND_FORMAT_FLOAT:
               fillOutputArray = SoundFileIn::fillOutputArray32F_B;
               break;
            case SND_FORMAT_DOUBLE:
               fillOutputArray = SoundFileIn::fillOutputArray64F_B;
               break;
            default:
               cerr << "Error: unknown input soundfile format: " 
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      case TYPE_WAV_PCM:                    // little endian samples
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               fillOutputArray = SoundFileIn::fillOutputArray8M_L;
               break;
            case SND_FORMAT_LINEAR_8:
               fillOutputArray = SoundFileIn::fillOutputArray8L_L;
               break;
            case SND_FORMAT_LINEAR_16:
               fillOutputArray = SoundFileIn::fillOutputArray16L_L;
               break;
            case SND_FORMAT_LINEAR_24:
               fillOutputArray = SoundFileIn::fillOutputArray24L_L;
               break;
            case SND_FORMAT_LINEAR_32:
               fillOutputArray = SoundFileIn::fillOutputArray32L_L;
               break;
            case SND_FORMAT_FLOAT:
               fillOutputArray = SoundFileIn::fillOutputArray32F_L;
               break;
            case SND_FORMAT_DOUBLE:
               fillOutputArray = SoundFileIn::fillOutputArray64F_L;
               break;
            default:
               cerr << "Error: unknown input soundfile format: " 
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      default:
         cerr << "Error: unknown soundfile type: "
              << sndHeader.getInputType() << endl;
         exit(1);
   }
}


//////////////////////////////
//
// SoundFileIn::fillOutputArray8M -- read channel samples as
//	8-bit Mulaw values
//

void SoundFileIn::fillOutputArray8M_B(void) {
   static int i;
   static uchar sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
      Output[i] = (sampleType)mulaw8ToLinear16(sample) / (long)0x8000;
   }
}


void SoundFileIn::fillOutputArray8M_L(void) {
   static int i;
   static uchar sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)mulaw8ToLinear16(sample) / (long)0x8000;
   }
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray8L
//

void SoundFileIn::fillOutputArray8L_B(void) {
   static int i;
   static char sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
      Output[i] = (sampleType)sample / (long)0x80;
   }
}


void SoundFileIn::fillOutputArray8L_L(void) {
   static int i;
   static char sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)sample / (long)0x80;
   }
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray16L
//
 
//#include <iomanip.h>

void SoundFileIn::fillOutputArray16L_B(void) {
   static int i;
   static short sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
//cerr << hex << sample << "";
      Output[i] = (sampleType)sample / (long)0x8000;
//cerr << "=" << Output[i] << endl;
   }
}


void SoundFileIn::fillOutputArray16L_L(void) {
   static int i;
   static short sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)sample / (long)0x8000;
   }
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray24L
//

void SoundFileIn::fillOutputArray24L_B(void) {
   // not yet complete
}

void SoundFileIn::fillOutputArray24L_L(void) {
   // not yet complete
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray32L
//

void SoundFileIn::fillOutputArray32L_B(void) {
   static int i;
   static long sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
      Output[i] = (sampleType)sample / (ulong)0x80000000;
   }
}


void SoundFileIn::fillOutputArray32L_L(void) {
   static int i;
   static long sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)sample / (ulong)0x80000000;
   }
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray32F
//

void SoundFileIn::fillOutputArray32F_B(void) {
   static int i;
   static float sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
      Output[i] = (sampleType)sample;
   }
}

void SoundFileIn::fillOutputArray32F_L(void) {
   static int i;
   static float sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)sample;
   }
}



//////////////////////////////
//
// SoundFileIn::fillOutputArray64F
//

void SoundFileIn::fillOutputArray64F_B(void) {
   static int i;
   static double sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readBigEndian(sample);
      Output[i] = (sampleType)sample;
   }
}


void SoundFileIn::fillOutputArray64F_L(void) {
   static int i;
   static double sample;

   for (i=0; i<sndHeader.getChannels(); i++) {
      sndfile.readLittleEndian(sample);
      Output[i] = (sampleType)sample;
   }
}



//////////////////////////////
//
// SoundFileIn::mulaw8ToLinear16 -- converts 8-bit mulaw to 
//	16-bit linear quantization
//

short SoundFileIn::mulaw8ToLinear16(uchar index) {
   static const short MuLaw[256] = {
   // negative values
    0x8284, 0x8684, 0x8a84, 0x8e84, 0x9284, 0x9684, 0x9a84, 0x9e84, 
    0xa284, 0xa684, 0xaa84, 0xae84, 0xb284, 0xb684, 0xba84, 0xbe84, 
    0xc184, 0xc384, 0xc584, 0xc784, 0xc984, 0xcb84, 0xcd84, 0xcf84, 
    0xd184, 0xd384, 0xd584, 0xd784, 0xd984, 0xdb84, 0xdd84, 0xdf84, 
    0xe104, 0xe204, 0xe304, 0xe404, 0xe504, 0xe604, 0xe704, 0xe804, 
    0xe904, 0xea04, 0xeb04, 0xec04, 0xed04, 0xee04, 0xef04, 0xf004, 
    0xf0c4, 0xf144, 0xf1c4, 0xf244, 0xf2c4, 0xf344, 0xf3c4, 0xf444, 
    0xf4c4, 0xf544, 0xf5c4, 0xf644, 0xf6c4, 0xf744, 0xf7c4, 0xf844, 
    0xf8a4, 0xf8e4, 0xf924, 0xf964, 0xf9a4, 0xf9e4, 0xfa24, 0xfa64, 
    0xfaa4, 0xfae4, 0xfb24, 0xfb64, 0xfba4, 0xfbe4, 0xfc24, 0xfc64, 
    0xfc94, 0xfcb4, 0xfcd4, 0xfcf4, 0xfd14, 0xfd34, 0xfd54, 0xfd74, 
    0xfd94, 0xfdb4, 0xfdd4, 0xfdf4, 0xfe14, 0xfe34, 0xfe54, 0xfe74, 
    0xfe8c, 0xfe9c, 0xfeac, 0xfebc, 0xfecc, 0xfedc, 0xfeec, 0xfefc, 
    0xff0c, 0xff1c, 0xff2c, 0xff3c, 0xff4c, 0xff5c, 0xff6c, 0xff7c, 
    0xff88, 0xff90, 0xff98, 0xffa0, 0xffa8, 0xffb0, 0xffb8, 0xffc0, 
    0xffc8, 0xffd0, 0xffd8, 0xffe0, 0xffe8, 0xfff0, 0xfff8, 0x0000, 
   // positive values
    0x7d7c, 0x797c, 0x757c, 0x717c, 0x6d7c, 0x697c, 0x657c, 0x617c, 
    0x5d7c, 0x597c, 0x557c, 0x517c, 0x4d7c, 0x497c, 0x457c, 0x417c, 
    0x3e7c, 0x3c7c, 0x3a7c, 0x387c, 0x367c, 0x347c, 0x327c, 0x307c, 
    0x2e7c, 0x2c7c, 0x2a7c, 0x287c, 0x267c, 0x247c, 0x227c, 0x207c, 
    0x1efc, 0x1dfc, 0x1cfc, 0x1bfc, 0x1afc, 0x19fc, 0x18fc, 0x17fc, 
    0x16fc, 0x15fc, 0x14fc, 0x13fc, 0x12fc, 0x11fc, 0x10fc, 0x0ffc, 
    0x0f3c, 0x0ebc, 0x0e3c, 0x0dbc, 0x0d3c, 0x0cbc, 0x0c3c, 0x0bbc, 
    0x0b3c, 0x0abc, 0x0a3c, 0x09bc, 0x093c, 0x08bc, 0x083c, 0x07bc, 
    0x075c, 0x071c, 0x06dc, 0x069c, 0x065c, 0x061c, 0x05dc, 0x059c, 
    0x055c, 0x051c, 0x04dc, 0x049c, 0x045c, 0x041c, 0x03dc, 0x039c, 
    0x036c, 0x034c, 0x032c, 0x030c, 0x02ec, 0x02cc, 0x02ac, 0x028c, 
    0x026c, 0x024c, 0x022c, 0x020c, 0x01ec, 0x01cc, 0x01ac, 0x018c, 
    0x0174, 0x0164, 0x0154, 0x0144, 0x0134, 0x0124, 0x0114, 0x0104, 
    0x00f4, 0x00e4, 0x00d4, 0x00c4, 0x00b4, 0x00a4, 0x0094, 0x0084, 
    0x0078, 0x0070, 0x0068, 0x0060, 0x0058, 0x0050, 0x0048, 0x0040, 
    0x0038, 0x0030, 0x0028, 0x0020, 0x0018, 0x0010, 0x0008, 0x0000
   };

   return MuLaw[index];
}


