//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Sat Nov  2 10:25:09 PST 1996
// Last Modified: Sat May 24 19:02:27 GMT-0800 1997
// Filename:      .../sig/Filter/SoundFileOut/SoundFileOut.cc
// Syntax:        C++ 
// $Smake:        cc -Wall -g -c %b.cc -I../../../include && rm -f %b.o
//

#include "SoundFileOut.h"
#include "HelperFunctions.h"
#include <stdlib.h>

#include <iostream.h>

//////////////////////////////
//
// SoundFileOut::SoundFileOut
//

SoundFileOut::SoundFileOut(char* aFileName, SoundHeader& header) {
   filename = new char[strlen(aFileName)+1];
   strcpy(filename, aFileName);
   outSample = 0;
   setName("SoundFileOut");
   outFile.open(filename, ios::out | ios::noreplace);
   if (!outFile.is_open()) {
      cerr << "Error: cannot write soundfile " << filename 
           << ". Perhaps the file already exists?" << endl;
      exit(1);
   }
   sndHeader = header;
   sndHeader.modifyFormat(aFileName);
   outFile << sndHeader;
   outFile.seekp(sndHeader.getDataByteOffset(), ios::beg);
   sampleCount = 0;
   brandname = FILTER;
   determineSampleWritingFunction();
}



//////////////////////////////
//
// SoundFileOut::~SoundFileOut
//


SoundFileOut::~SoundFileOut() {
   close();
}



//////////////////////////////
//
// SoundFileOut::action
//

void SoundFileOut::action(void) {
   for (int i=0; i<sndHeader.getChannels(); i++) {
      writeSample(inputs[i]);
   }
   sampleCount++;
}



//////////////////////////////
//
// SoundFileOut::close
//

void SoundFileOut::close(void) {
   if (sndHeader.getSamples() != sampleCount) {
      switch (sndHeader.getOutputType()) {
         case TYPE_SND:
            // write the number of bytes in the data:
            outFile.seekp(8, ios::beg);   
            outFile.writeBigEndian((long)
               (sampleCount*sndHeader.getBlockAlign()));
            break;
         case TYPE_WAV_PCM:
            outFile.seekp(8, ios::beg);
            // write the number of bytes in the file minus a few:
            outFile.seekp(4, ios::beg);   
            outFile.writeLittleEndian((long)
               (36 + sampleCount * sndHeader.getBlockAlign()));
            // write the number of sound data bytes:
            outFile.seekp(40, ios::beg);
            outFile.writeLittleEndian((long)
               (sampleCount * sndHeader.getBlockAlign()));
            break;
         default:
           cerr << "Unknown output soundfile type: "
                << sndHeader.getOutputType() << endl;
           exit(1);
      }
   }
   outFile.close();
   sampleCount = 0;
}



//////////////////////////////
//
// SoundFileOut::output
//

sampleType SoundFileOut::output(int channel) {
   return inputs[channel];
}



///////////////////////////////////////////////////////////////////////////
//
// private functions:
//


//////////////////////////////
//
// SoundFileIn::determineSampleWritingFunction
//

void SoundFileOut::determineSampleWritingFunction(void) {
   switch (sndHeader.getOutputType()) {
      case TYPE_SND:                           // bigendian samples
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               writeSample = SoundFileOut::writeSample8M_B;
               break;
            case SND_FORMAT_LINEAR_8:
               writeSample = SoundFileOut::writeSample8L_B;
               break;
            case SND_FORMAT_LINEAR_16:
               writeSample = SoundFileOut::writeSample16L_B;
               break;
            case SND_FORMAT_LINEAR_24:
               writeSample = SoundFileOut::writeSample24L_B;
               break;
            case SND_FORMAT_LINEAR_32:
               writeSample = SoundFileOut::writeSample32L_B;
               break;
            case SND_FORMAT_FLOAT:
               writeSample = SoundFileOut::writeSample32F_B;
               break;
            case SND_FORMAT_DOUBLE:
               writeSample = SoundFileOut::writeSample64F_B;
               break;
            default:
               cerr << "Error: unknown output soundfile format: "
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      case TYPE_WAV_PCM:                      // little endian samples
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               writeSample = SoundFileOut::writeSample8M_L;
               break;
            case SND_FORMAT_LINEAR_8:
               writeSample = SoundFileOut::writeSample8L_L;
               break;
            case SND_FORMAT_LINEAR_16:
               writeSample = SoundFileOut::writeSample16L_L;
               break;
            case SND_FORMAT_LINEAR_24:
               writeSample = SoundFileOut::writeSample24L_L;
               break;
            case SND_FORMAT_LINEAR_32:
               writeSample = SoundFileOut::writeSample32L_L;
               break;
            case SND_FORMAT_FLOAT:
               writeSample = SoundFileOut::writeSample32F_L;
               break;
            case SND_FORMAT_DOUBLE:
               writeSample = SoundFileOut::writeSample64F_L;
               break;
            default:
               cerr << "Error: unknown output soundfile format: "
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      default:
         cerr << "Error: unknown soundfile type: "
              << sndHeader.getOutputType() << endl;
         exit(1);
   }
}



//////////////////////////////
//
// SoundFileOut::writeSample8M -- read channel samples as
//      8-bit Mulaw values
//

void SoundFileOut::writeSample8M_B(sampleType aSample) {
   outFile.writeBigEndian(linear16ToMulaw8(SampleToShort(aSample)));
}


void SoundFileOut::writeSample8M_L(sampleType aSample) {
   outFile.writeLittleEndian(linear16ToMulaw8(SampleToShort(aSample)));
}



//////////////////////////////
//
// SoundFileOut::writeSample8L
//

void SoundFileOut::writeSample8L_B(sampleType aSample) {
   outFile.writeBigEndian(SampleToChar(aSample));
}


void SoundFileOut::writeSample8L_L(sampleType aSample) {
   outFile.writeBigEndian(SampleToChar(aSample));
}



//////////////////////////////
//
// SoundFileOut::writeSample16L
//

void SoundFileOut::writeSample16L_B(sampleType aSample) {
   outFile.writeBigEndian(SampleToShort(aSample));
}


void SoundFileOut::writeSample16L_L(sampleType aSample) {
   outFile.writeLittleEndian(SampleToShort(aSample));
}



//////////////////////////////
//
// SoundFileOut::writeSample24L
//

void SoundFileOut::writeSample24L_B(sampleType aSample) {
   static long output;
   output = SampleToLong(aSample);
   outFile.writeBigEndian((uchar)(output >> 12));
   outFile.writeBigEndian((uchar)(output >>  8));
   outFile.writeBigEndian((uchar)(output >>  4));
}


void SoundFileOut::writeSample24L_L(sampleType aSample) {
   static long output;
   output = SampleToLong(aSample);
   outFile.writeBigEndian((uchar)(output >>  4));
   outFile.writeBigEndian((uchar)(output >>  8));
   outFile.writeBigEndian((uchar)(output >> 12));
}



//////////////////////////////
//
// SoundFileOut::writeSample32L
//

void SoundFileOut::writeSample32L_B(sampleType aSample) {
   outFile.writeBigEndian(SampleToLong(aSample));
}


void SoundFileOut::writeSample32L_L(sampleType aSample) {
   outFile.writeBigEndian(SampleToLong(aSample));
}



//////////////////////////////
//
// SoundFileOut::writeSample32F
//

void SoundFileOut::writeSample32F_B(sampleType aSample) {
   outFile.writeBigEndian((float)aSample);
}


void SoundFileOut::writeSample32F_L(sampleType aSample) {
   outFile.writeLittleEndian((float)aSample);
}



//////////////////////////////
//
// SoundFileOut::writeSample64F
//

void SoundFileOut::writeSample64F_B(sampleType aSample) {
   outFile.writeBigEndian((double)aSample);
}


void SoundFileOut::writeSample64F_L(sampleType aSample) {
   outFile.writeLittleEndian((double)aSample);
}



//////////////////////////////
//
// SoundFileOut::linear16ToMulaw8
//
// reference: http://www.omnigroup.com/MailArchive/next-prog/1994/0042.html
//

#define BIAS 0x84 /* define the add-in bias for 16 bit samples */
#define CLIP 32635

uchar SoundFileOut::linear16ToMulaw8(short aSample) {

   static int exponentLookupTable[256] = {
      0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
      4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7
   };

   static int sign, exponent, mantissa, magnitude;
   static uchar output;

   // put the sample into sign-magnitude form
   sign = (aSample >> 8) & 0x80;                  // set aside the sign 
   magnitude = sign ? -aSample : aSample;         // get magnitude
   if (magnitude > CLIP) magnitude = CLIP;        // clip the magnitude 

   // Convert from 16 bit linear to ulaw. 
   magnitude += BIAS;
   exponent   = exponentLookupTable[(uchar)((magnitude >> 7) & 0xFF)];
   mantissa   = (magnitude >> (exponent + 3)) & 0x0F;
   output     = ~(sign | (exponent << 4) | mantissa);
   
   return output;
}



